/*
 * thread-tracker.c - Tracks threads in a GstPipeline
 * Copyright (C) 2009-2010 Collabora Ltd.
 * Copyright (C) 2009-2010 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "thread-tracker.h"

#include <pthread.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <gst/gst.h>

static gboolean add_pad_locked (struct ObjectTrackingData *otd, GstPad *pad);


void
thread_tracking_init (struct ThreadTrackingData *ttd,
    GMutex *mutex,
    setup_thread_prio restore_thread_prio)
{
  ttd->object_threads =
      g_hash_table_new (g_direct_hash, g_direct_equal);
  ttd->object_tids =
      g_hash_table_new (g_direct_hash, g_direct_equal);
  ttd->otds = g_ptr_array_new ();
  ttd->mutex = mutex;
  ttd->restore_thread_prio = restore_thread_prio;
}

void
object_tracking_init (struct ThreadTrackingData *ttd,
    struct ObjectTrackingData *otd,
    setup_thread_prio setup_thread_prio)
{
  otd->ttd = ttd;
  otd->objects = g_ptr_array_new ();
  otd->setup_thread_prio = setup_thread_prio;
  g_ptr_array_add (ttd->otds, otd);
}


void
thread_tracking_data_free (struct ThreadTrackingData *ttd)
{
  g_hash_table_destroy (ttd->object_threads);
  g_hash_table_destroy (ttd->object_tids);
  g_ptr_array_free (ttd->otds, TRUE);
}

void
object_tracking_data_free (struct ObjectTrackingData *otd)
{
  g_ptr_array_free (otd->objects, TRUE);
  g_ptr_array_remove (otd->ttd->otds, otd);
}



static void
remove_object (gpointer data, GObject *where_the_obj_was)
{
  struct ObjectTrackingData *otd = data;

  g_mutex_lock (otd->ttd->mutex);
  g_ptr_array_remove (otd->objects, where_the_obj_was);
  g_mutex_unlock (otd->ttd->mutex);
}

static void
pad_linked (GstPad *pad, GstPad *peer, struct ObjectTrackingData *otd);

/* Returns TRUE if it is a object
 * Returns FALSE if it was already added
 */

static gboolean
insert_object_locked (struct ObjectTrackingData *otd, gpointer object)
{
  guint i;
  pthread_t thread;
  pid_t tid;

  for (i = 0; i < otd->objects->len; i++)
    if (g_ptr_array_index (otd->objects, i) == object)
      return FALSE;

  g_ptr_array_add (otd->objects, object);
  g_object_weak_ref (G_OBJECT (object), remove_object, otd);


  thread = GPOINTER_TO_UINT (
      g_hash_table_lookup (otd->ttd->object_threads, object));
  tid = GPOINTER_TO_UINT (
      g_hash_table_lookup (otd->ttd->object_tids, object));

  if (thread && tid)
    otd->setup_thread_prio (thread, tid);

  return TRUE;
}

static void
pad_added (GstElement *element, GstPad *pad, struct ObjectTrackingData *otd)
{
  add_pad_locked (otd, pad);
}

/* Returns TRUE if it is a new pad
 * Returns FALSE if it was already added
 */

static gboolean
add_pad_locked (struct ObjectTrackingData *otd, GstPad *pad)
{
  GstObject *parent;

  if (!insert_object_locked (otd, pad))
    return FALSE;

  parent = gst_pad_get_parent (pad);

  if (parent)
    {
      if (GST_IS_PAD (parent))
        {
          add_pad_locked (otd, GST_PAD_CAST (parent));
        }
      else
        {
          if (insert_object_locked (otd, parent) &&
              (GST_IS_ELEMENT (parent) && !GST_IS_BIN (parent)))
            {
              g_signal_connect (parent, "pad-added", G_CALLBACK (pad_added),
                  otd);
            }
        }
      gst_object_unref (parent);
    }

  g_signal_connect (pad, "linked", G_CALLBACK (pad_linked), otd);

  return TRUE;
}

static void
pad_iter_locked (gpointer data, gpointer user_data)
{
  struct ObjectTrackingData *otd = user_data;
  GstPad *pad = data;

  if (add_pad_locked (otd, pad))
    {
      GstPad *peerpad = gst_pad_get_peer (pad);

      if (peerpad)
        {
          if (add_pad_locked (otd, peerpad))
            {
              GstIterator *iter;

              iter = gst_pad_iterate_internal_links (peerpad);
              if (iter)
                {
                  while (gst_iterator_foreach (iter, pad_iter_locked, otd) ==
                      GST_ITERATOR_RESYNC)
                    gst_iterator_resync (iter);
                  gst_iterator_free (iter);
                }
            }
          gst_object_unref (peerpad);
        }
    }
  gst_object_unref (pad);
}

static void
pad_linked (GstPad *pad, GstPad *peer, struct ObjectTrackingData *otd)
{
  g_mutex_lock (otd->ttd->mutex);
  if (add_pad_locked (otd, peer))
    {
      GstIterator *iter = gst_pad_iterate_internal_links (peer);

      if (iter)
        {
          while (gst_iterator_foreach (iter, pad_iter_locked, otd) ==
              GST_ITERATOR_RESYNC)
            gst_iterator_resync (iter);

          gst_iterator_free (iter);
        }
    }
  g_mutex_unlock (otd->ttd->mutex);
}

static void
pad_internal_links_iter_locked (gpointer data, gpointer user_data)
{
  struct ObjectTrackingData *otd = user_data;
  GstPad *pad = data;
  GstIterator *iter;

  iter = gst_pad_iterate_internal_links (pad);
  while (gst_iterator_foreach (iter, pad_iter_locked, otd) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (iter);
  gst_iterator_free (iter);

  gst_object_unref (pad);
}


void
object_tracker_add_sink (struct ObjectTrackingData *otd, GstElement *sink)
{
  GstIterator *iter;

  g_mutex_lock (otd->ttd->mutex);

  iter = gst_element_iterate_sink_pads (sink);
  while (gst_iterator_foreach (iter, pad_iter_locked, otd) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (iter);

  gst_iterator_resync (iter);
  while (gst_iterator_foreach (iter, pad_internal_links_iter_locked, otd) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (iter);
  gst_iterator_free (iter);

  g_mutex_unlock (otd->ttd->mutex);
}

void
object_tracker_add_src (struct ObjectTrackingData *otd, GstElement *src)
{
  GstIterator *iter;

  g_mutex_lock (otd->ttd->mutex);

  iter = gst_element_iterate_src_pads (src);
  while (gst_iterator_foreach (iter, pad_iter_locked, otd) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (iter);

  gst_iterator_resync (iter);
  while (gst_iterator_foreach (iter, pad_internal_links_iter_locked, otd) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (iter);
  gst_iterator_free (iter);

  g_mutex_unlock (otd->ttd->mutex);
}


void
enter_thread (struct ThreadTrackingData *ttd, GstObject *src)
{
  pthread_t thread = pthread_self ();
  pid_t tid = syscall (__NR_gettid);
  guint i, j;

  g_mutex_lock (ttd->mutex);
  g_hash_table_insert (ttd->object_threads, src, GINT_TO_POINTER (thread));
  g_hash_table_insert (ttd->object_tids, src, GINT_TO_POINTER (tid));

  for (i=0; i < ttd->otds->len; i++)
    {
      struct ObjectTrackingData *otd = g_ptr_array_index (ttd->otds, i);

      for (j = 0; j < otd->objects->len; j++)
        {
          if (g_ptr_array_index (otd->objects, j) == src)
            otd->setup_thread_prio (thread, tid);
          break;
        }
    }

  g_mutex_unlock (ttd->mutex);
}


void
leave_thread (struct ThreadTrackingData *ttd, GstObject *src, GstElement *owner)
{
  pthread_t thread = pthread_self ();
  pid_t tid = syscall (__NR_gettid);

  g_mutex_lock (ttd->mutex);
  g_hash_table_remove (ttd->object_threads, src);
  g_hash_table_remove (ttd->object_tids, src);
  g_mutex_unlock (ttd->mutex);

  ttd->restore_thread_prio (thread, tid);
}
