/*
 * videosink.c - Source for TpStreamEngineVideoSink
 * Copyright (C) 2006-2008 Collabora Ltd.
 * Copyright (C) 2006-2008 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "videosink.h"

#include <string.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <gst/interfaces/meegovideotexture.h>

#include <gst/interfaces/xoverlay.h>
#include <gst/video/video.h>
#include <gst/farsight/fs-element-added-notifier.h>

#include <tp-stream-engine.h>
#include <mqueue.h>
#define QMQ_PRIO        (MQ_PRIO_MAX - 1)

#include "tp-stream-engine-signals-marshal.h"
#include "util.h"
//#include "extdefs.h"
//_PFNEGLCREATESYNCKHRPROC eglCreateSyncKHR;

#define DEFAULT_XV_DEVICE "0"

// mq ipc events, shared between SE and call-ui
#define CONNECT_MSG 100
#define REPAINT_MSG 101
#define IPC_MQ_PREVIEW_CLIENT "/mq-previewClient"
#define IPC_MQ_REMOTE_CLIENT "/mq-remoteClient"
#define IPC_MQ_PREVIEW_SERVER "/mq-previewVideo"
#define IPC_MQ_REMOTE_SERVER "/mq-remoteVideo"



G_DEFINE_ABSTRACT_TYPE (TpStreamEngineVideoSink, tp_stream_engine_video_sink,
    G_TYPE_OBJECT);

struct _TpStreamEngineVideoSinkPrivate
{
  GstElement *sink;
  GstElement *glsink;

  GstPad *sinkpad;

  guint window_id;
  WindowParams win_params;
  gboolean is_preview;

  GstClockTime last_timestamp;
  gdouble framerate;
  GstElement *textoverlay;

  gulong caps_changed_id;

  GStaticMutex mutex;
  guint size_change_idle_id; /* mutex protected */

  guint process_video_frame_id;
  GStaticMutex gl_sink_mutex;
  int frame_id;
  GLint shader_handle;
  int frame_ready_id;

  Display *xdpy;
  EGLDisplay egl_display;
  EGLContext egl_context;
  EGLSurface egl_win_surface;

  mqd_t mq_client;
  mqd_t mq_server;
  uint pixmap_handle;
  int width;
  int height;
};

/* properties */
enum
{
  PROP_0,
  PROP_SINK,
  PROP_WINDOW_ID,
  PROP_WINDOW_PARAMS,
  PROP_IS_PREVIEW,
  PROP_XV_DEVICE
};


/* signals */

enum
{
  PLUG_DELETED,
  SIZE_CHANGED,
  SIGNAL_COUNT
};


static guint signals[SIGNAL_COUNT] = {0};

#if !defined(GL_TEXTURE_EXTERNAL_OES)
#define GL_TEXTURE_EXTERNAL_OES 0x8D65
#endif

const char* fragment_shader = ""
"#extension GL_OES_EGL_image_external: enable\n"
"\n"
"uniform samplerExternalOES texture0;\n"
"varying lowp vec2 myTexCoord;\n"
"\n"
"void main (void)\n"
"{\n"
"        gl_FragColor = texture2D(texture0, myTexCoord);\n"
"}";

const char *vertex_shader = ""
"attribute highp vec4 vertex;"
"attribute mediump vec4  myUV;"
"varying lowp vec2 myTexCoord;"
"void main (void)"
"{"
"gl_Position = vertex;"
"myTexCoord = myUV.st;"
"}";

typedef struct
{
  unsigned int id;
} Msg;

void
create_ipc (TpStreamEngineVideoSink *self, gboolean is_preview)
{
  int flags;
  struct mq_attr attr;
  gchar *client_path;
  gchar *server_path;
  Msg message;

  if (is_preview)
    {
      client_path = g_strdup(IPC_MQ_PREVIEW_CLIENT);
      server_path = g_strdup(IPC_MQ_PREVIEW_SERVER);
    }
  else
    {
      client_path = g_strdup(IPC_MQ_REMOTE_CLIENT);
      server_path = g_strdup(IPC_MQ_REMOTE_SERVER);
    }

  if (self->priv->mq_client != ((mqd_t) -1))
    {
      if (mq_unlink(client_path) != 0)
        {
          if (errno != ENOENT)
              g_debug("unable to unlink client path");
        }
      if (mq_close(self->priv->mq_client) != 0)
        g_debug("unable  to close client mqueue");
    }

  flags = O_RDWR | O_CREAT;
  mode_t mode  = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;

  attr.mq_flags   = 0;
  attr.mq_maxmsg  = 10;
  attr.mq_msgsize = sizeof(Msg);
  attr.mq_curmsgs = 0;

  self->priv->mq_client = mq_open(client_path, flags, mode, &attr);
  if (self->priv->mq_client == (mqd_t) -1)
    g_debug("Error creating client mqueue: %d",errno);

  flags = O_WRONLY;
  flags |= O_NONBLOCK;

  self->priv->mq_server = mq_open(server_path, flags);
  if (self->priv->mq_server == (mqd_t) -1)
    g_debug("Error opening the server mqueue %d", errno);

  message.id = CONNECT_MSG; //connect message
  if (mq_send (self->priv->mq_server, (char*) &message,
      sizeof (Msg), QMQ_PRIO) == -1)
    {
      g_debug ("Error sending msg %d", errno);
    }
  g_free (client_path);
  g_free (server_path);
}

void
send_msg (TpStreamEngineVideoSink *self, int id)
{
  if (self->priv->mq_server == (mqd_t) -1)
    {
      g_debug("No server mqueue");
      return;
    }
  Msg msg;
  msg.id = id;
  //g_debug("Sending msg to mqueue %d ", msg.id);
  if (mq_send(self->priv->mq_server, (char*) &msg, sizeof(Msg),0) == -1)
    {
      g_debug("Error sending msg %d",errno);
    }
}

void
close_ipc (TpStreamEngineVideoSink *self)
{
  g_debug ("close_ipc");
  if (self->priv->mq_server != ((mqd_t) -1))
    mq_close (self->priv->mq_server);
  if (self->priv->mq_client != ((mqd_t) -1))
    {
      mq_close (self->priv->mq_client);
      if (self->priv->is_preview)
        mq_unlink (IPC_MQ_PREVIEW_CLIENT);
      else
        mq_unlink (IPC_MQ_REMOTE_CLIENT);
    }
}

Display * get_x_display()
{
  static Display *xdpy = NULL;
  if (xdpy == NULL)
    xdpy = XOpenDisplay (NULL);
  return xdpy;
}

GLint
create_shader_program (const char* vertex_shader_string,
    const char* fragment_shader_string)
{
  g_debug("create_shader_program");
  GLint program_handle, fragment_shader_handle, vertex_shader_handle;

  fragment_shader_handle = glCreateShader (GL_FRAGMENT_SHADER);
  glShaderSource(fragment_shader_handle, 1, &fragment_shader_string, NULL);
  glCompileShader(fragment_shader_handle);

  vertex_shader_handle   = glCreateShader (GL_VERTEX_SHADER);
  glShaderSource(vertex_shader_handle, 1, &vertex_shader_string, NULL);
  glCompileShader(vertex_shader_handle);

  program_handle = glCreateProgram();

  glAttachShader(program_handle, vertex_shader_handle);
  glAttachShader(program_handle, fragment_shader_handle);

  return program_handle;
}

static void
tp_stream_engine_video_sink_init (TpStreamEngineVideoSink *self)
{
  g_debug("video_sink_init");
  TpStreamEngineVideoSinkPrivate *priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
      TP_STREAM_ENGINE_TYPE_VIDEO_SINK, TpStreamEngineVideoSinkPrivate);

  self->priv = priv;
  self->priv->shader_handle = 0;
  self->priv->pixmap_handle = 0;
  self->priv->frame_id = -1;
  self->priv->process_video_frame_id = 0;
  self->priv->mq_client = (mqd_t) -1;
  self->priv->mq_server = (mqd_t) -1;
  self->priv->width = 320;
  self->priv->height = 240;
  g_static_mutex_init (&priv->mutex);
  g_static_mutex_init (&priv->gl_sink_mutex);
  priv->last_timestamp = GST_CLOCK_TIME_NONE;
}


static gboolean
fps_qos_probe (GstPad *pad, GstEvent *event, gpointer user_data)
{
  TpStreamEngineVideoSink *self = TP_STREAM_ENGINE_VIDEO_SINK (user_data);
  gdouble proportion;
  GstClockTimeDiff diff;
  GstClockTime timestamp;

  if (GST_EVENT_TYPE (event) != GST_EVENT_QOS)
    return TRUE;

  gst_event_parse_qos (event, &proportion, &diff, &timestamp);

  /* Ignore late buffers */
  if (diff > 0)
    return TRUE;

  if (GST_CLOCK_TIME_IS_VALID (self->priv->last_timestamp))
    {
      GstClockTimeDiff diff = GST_CLOCK_DIFF(self->priv->last_timestamp,
          timestamp);
      gdouble diffs;

      if (diff > GST_SECOND)
        diff = GST_SECOND;

      diffs = (gdouble)diff / (gdouble)GST_SECOND;
      if (self->priv->framerate != 0)
        {
        self->priv->framerate = (1.0/diffs)*(diffs) + self->priv->framerate*(1-diffs);

        if (self->priv->textoverlay)
          {
            gchar *str = g_strdup_printf ("FPS: %f", self->priv->framerate);
            g_object_set (self->priv->textoverlay, "text", str,
                "halignment", 0,
                "valignment", 2,
                "auto-resize", FALSE, NULL);
            g_free (str);
          }
        else
          {
            g_debug("%s:%s: Last diff: %" G_GINT64_FORMAT " diffs: %f"
                " Instant fps: %f averaged over 1s: %f\n",
                GST_DEBUG_PAD_NAME (pad), diff,
                diffs, (1.0/diffs), self->priv->framerate);
          }
        }
      else
        {
          self->priv->framerate = 1.0/diffs;
        }
    }

  self->priv->last_timestamp = timestamp;

  return TRUE;
}

static gboolean
size_change_idle_cb (gpointer data)
{
  TpStreamEngineVideoSink *self = data;
  gint width, height;
  gboolean have_size = FALSE;
  g_debug("size_change_idle_cb");

  g_static_mutex_lock (&self->priv->mutex);
  self->priv->size_change_idle_id = 0;
  g_static_mutex_unlock (&self->priv->mutex);

  GST_OBJECT_LOCK (self->priv->sinkpad);
  if (gst_video_get_size (self->priv->sinkpad, &width, &height))
    {
      gint par_n, par_d;

      if (gst_video_parse_caps_pixel_aspect_ratio (
              GST_PAD_CAPS (self->priv->sinkpad), &par_n, &par_d))
        {
          g_debug (" size_change_idle_cb width :%d height: %d par_n: %d par_d: %d",
              width, height, par_n, par_d); 
          width *= par_n;
          width /= par_d;
        }
      have_size = TRUE;
    }
  GST_OBJECT_UNLOCK (self->priv->sinkpad);

  if (have_size)
    g_signal_emit (self, signals[SIZE_CHANGED], 0, width, height);

  return FALSE;
}

static void
caps_changed (GObject *obj, GParamSpec *spec, gpointer user_data)
{
  TpStreamEngineVideoSink *self = user_data;
  g_debug("caps_changed");
  gint width, height;
  g_static_mutex_lock (&self->priv->mutex);
  if (gst_video_get_size (self->priv->sinkpad, &width, &height))
    {
      g_debug("Video width = %d Height = %d",width, height);
      self->priv->width = width;
      self->priv->height = height;
    }
  else
    {
      g_debug("Coudlnt get the video dimensions, going to fail safe size");
      self->priv->width = 320;
      self->priv->height = 240;
    }
  if (!self->priv->size_change_idle_id)
    self->priv->size_change_idle_id =
        g_idle_add ((GSourceFunc) size_change_idle_cb, self);
  g_static_mutex_unlock (&self->priv->mutex);
}


void
create_video_buffer(TpStreamEngineVideoSink *self,int width, int height)
{
  g_debug("create_video_buffer : Width: %d Height: %d self->priv : %p",
            width, height, self->priv);
  if (width <= 0 || height <= 0)
     return;
     
  Display *xdpy = get_x_display();
  EGLDisplay egl_display = eglGetDisplay((EGLNativeDisplayType)(xdpy));

  self->priv->pixmap_handle = XCreatePixmap( self->priv->xdpy,
      RootWindow (self->priv->xdpy,DefaultScreen (self->priv->xdpy)),
      width,
      height,
      XDefaultDepthOfScreen(DefaultScreenOfDisplay(self->priv->xdpy)) );
  g_debug("X pixmap handle %d", self->priv->pixmap_handle);

  EGLContext context = 0;
  EGLSurface eglPixmapSurface = 0;
  static const int MaxConfigs = 16;
  EGLint numConfigs;
  EGLConfig configs[MaxConfigs];
  const EGLint configAttributes[] = { EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT, EGL_NONE };

  eglChooseConfig( egl_display,
      configAttributes,
      configs,
      MaxConfigs,
      &numConfigs);
  for (int i = 0; i < numConfigs; ++i)
    {
       const EGLConfig config = configs[i];
       const EGLint contextAttributes[] = { EGL_CONTEXT_CLIENT_VERSION, 2, EGL_NONE };
       context = eglCreateContext(egl_display,
                    config, NULL, contextAttributes);
       if (!context)
          continue;

       eglPixmapSurface = eglCreatePixmapSurface (egl_display,
                              config,
                              (EGLNativePixmapType)(self->priv->pixmap_handle),
                              0);
       if (eglPixmapSurface)
           break;
       eglDestroyContext(self->priv->egl_display, context);
    }

  self->priv->egl_win_surface = eglPixmapSurface;
  self->priv->egl_context = context;
  g_debug ("egl_win_surface: %p egl_context: %p",self->priv->egl_win_surface,
      self->priv->egl_context);

  g_object_set( G_OBJECT(self->priv->glsink),
          "egl-context",
          self->priv->egl_context,
          NULL);

  create_ipc(self, self->priv->is_preview);
  g_debug ("server ipc queue: %d, client ipc queue: %d",
      self->priv->mq_server, self->priv->mq_client);
  XSync (xdpy, FALSE);
  send_msg( self, self->priv->pixmap_handle );
}


void render_frame(TpStreamEngineVideoSink *self)
{
  //self->priv->process_video_frame_id = 0;
  if (self->priv->frame_id == -1)
    {
      g_debug("frame_id = -1");
      return;
    }
  if (self->priv->pixmap_handle == 0)
    {
      g_debug("pixmap_handle %d",self->priv->pixmap_handle);
      create_video_buffer( self, self->priv->width, self->priv->height);
    }

  if (eglMakeCurrent (self->priv->egl_display,
          self->priv->egl_win_surface,
          self->priv->egl_win_surface,
          self->priv->egl_context) == EGL_TRUE)
    {
      /*g_object_set( G_OBJECT(self->priv->glsink),
          "egl-context",
          self->priv->egl_context,
          NULL);*/
    }
  else
    {
      g_debug("failed to make  egl context as current in rendering thread:");
      return;
    }

  eglBindAPI(EGL_OPENGL_API);

  static float quad[] = {
    -1.0,  1.0,
    1.0,  1.0,
    -1.0, -1.0,
    1.0, -1.0
  };

  static const GLfloat texCoords[] = {
    0., 0.,
    1., 0.,
    0., 1.,
    1., 1.,
  };

  if (!self->priv->shader_handle) {
    self->priv->shader_handle = create_shader_program (vertex_shader,
        fragment_shader);
    if (!self->priv->shader_handle)
      {
        g_debug("Failed to create shader");
        return;
      }
    glBindAttribLocation(self->priv->shader_handle, 1, "myUV");
    glLinkProgram(self->priv->shader_handle);
    glUseProgram(self->priv->shader_handle);
    // Set the sampler2D uniforms to corresponding texture units.
    glUniform1i(glGetUniformLocation(self->priv->shader_handle, "texture0"), 0);
  }

  glUseProgram(self->priv->shader_handle);
  glViewport(0, 0, self->priv->width, self->priv->height);
  glClearColor(0, 0, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT);

  if (!meego_gst_video_texture_acquire_frame (
      (MeegoGstVideoTexture*)(self->priv->glsink),
      self->priv->frame_id))
    {
       g_debug ("application failed to acquire the frame..");
       return;
    }

  glEnableVertexAttribArray(0);
  glEnableVertexAttribArray(1);
  glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, quad);
  glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, texCoords);

  if (eglGetCurrentContext() == EGL_NO_CONTEXT)
    {
      if (eglMakeCurrent (self->priv->egl_display,
                          self->priv->egl_win_surface,
                          self->priv->egl_win_surface,
                          self->priv->egl_context) == EGL_TRUE)
        {
          g_debug("make egl context as  current in rendering thread: \n");
        }
      else
        {
          g_debug ("failed to make  egl context as current in rendering thread: \n");
        }
    }

  if (!meego_gst_video_texture_bind_frame((MeegoGstVideoTexture*)(self->priv->glsink),
          GL_TEXTURE_EXTERNAL_OES, self->priv->frame_id))
    {
      g_debug ("application failed to  bind with the frame ");
      return;
    }

  glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

  if (!meego_gst_video_texture_bind_frame((MeegoGstVideoTexture*)(self->priv->glsink),
          GL_TEXTURE_EXTERNAL_OES, -1))
    {
      g_debug("application failed to  bind with the frame -1 ");
      return;
    }

  glUseProgram(0);

  glFlush();
  glFinish();

  //EGLSyncKHR sync = eglCreateSyncKHR(self->priv->egl_display, EGL_SYNC_FENCE_KHR, NULL);

  //eglSwapBuffers(self->priv->egl_display, self->priv->egl_win_surface);

  meego_gst_video_texture_release_frame((MeegoGstVideoTexture*)(self->priv->glsink),
      self->priv->frame_id, sync);

  self->priv->frame_id = -1;
  eglMakeCurrent(self->priv->egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
}

static
gboolean process_video_frame (gpointer data)
{
  TpStreamEngineVideoSink *self = data;
  //g_debug("process_video_frame");

  g_static_mutex_lock( &self->priv->gl_sink_mutex );
  render_frame( self );
  g_static_mutex_unlock( &self->priv->gl_sink_mutex );
  // send msg to UI to repaint
  send_msg( self, REPAINT_MSG ); // repaint message
  return FALSE;
}

void
frame_ready_handler (MeegoGstVideoTexture *sink, gint frame, gpointer data)
{
  TpStreamEngineVideoSink *self = data;
 // g_debug("frame_ready_handler frame: %d, sink:%p self->priv: %p",
    //  frame, sink, self->priv);

  g_static_mutex_lock (&self->priv->gl_sink_mutex);
  self->priv->frame_id = frame;
  g_static_mutex_unlock (&self->priv->gl_sink_mutex);

  process_video_frame(data);

  /*
  if (self->priv->process_video_frame_id == 0)
  {
    self->priv->process_video_frame_id = g_timeout_add_full(G_PRIORITY_HIGH,
        0, process_video_frame, (gpointer)self, 0);
    //g_debug("process_video_frame_id: %d", self->priv->process_video_frame_id);
  }
  g_static_mutex_unlock (&self->priv->gl_sink_mutex);*/
}

static GstElement *
make_video_sink (TpStreamEngineVideoSink *self, gboolean is_preview)
{
  g_debug("make video sink element self->priv: %p, is preview? %d",
      self->priv, is_preview);
  const gchar *videosink_name;
  GstElement *sink = NULL;
  GstElement *bin, *tmp;
  GstPad *ghostpad;

  self->priv->xdpy = get_x_display();
  g_debug ("xdisplay %p", self->priv->xdpy);
  self->priv->egl_display = eglGetDisplay((EGLNativeDisplayType) self->priv->xdpy );
  EGLint major, minor;
  if (!eglInitialize(self->priv->egl_display, &major, &minor))
    {
      g_debug("error, could not init display");
    }

  sink = gst_element_factory_make("gltexturesink", NULL);
  if (sink == NULL)
    {
      g_debug ("failed to make a video sink");
      return NULL;
    }
  self->priv->glsink = sink;
  g_object_set(G_OBJECT(sink), "egl-display", self->priv->egl_display,NULL);
  g_object_set(G_OBJECT(sink), "x-display", self->priv->xdpy, NULL);
  g_object_set(G_OBJECT(sink), "use-framebuffer-memory", TRUE, NULL);
  g_object_set(G_OBJECT(sink), "render-mode", "video-texture", NULL);
  self->priv->frame_ready_id = g_signal_connect(G_OBJECT(sink), "frame-ready",
                                         G_CALLBACK(frame_ready_handler), self);

  g_debug ("made gltexture video sink element %s", GST_ELEMENT_NAME (sink));
  eglMakeCurrent (self->priv->egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
  // EGLSync functions
  //eglCreateSyncKHR = (_PFNEGLCREATESYNCKHRPROC)eglGetProcAddress("eglCreateSyncKHR");

  bin = gst_bin_new (NULL);
  if (!gst_bin_add (GST_BIN (bin), sink))
    {
      g_warning ("Could not add sink to the sinkbin");
      gst_object_unref (bin);
      gst_object_unref (sink);
      return NULL;
    }

  if (g_getenv ("STREAM_ENGINE_FPS") &&
      !strcmp (g_getenv ("STREAM_ENGINE_FPS"), "overlay"))
    {
      tmp = gst_element_factory_make ("textoverlay", NULL);

      if (tmp != NULL)
        {
          self->priv->textoverlay = tmp;
          g_debug ("linking textoverlay");
          if (!gst_bin_add (GST_BIN (bin), tmp))
            {
              g_warning ("Could not add textoverlay to the source bin");
              gst_object_unref (tmp);
              gst_object_unref (bin);
              return NULL;
            }
          if (!gst_element_link (tmp, sink))
            {
              g_warning ("Could not link sink and textoverlay elements");
              gst_object_unref (bin);
              return NULL;
            }
          sink = tmp;
          self->priv->sinkpad = gst_element_get_static_pad (tmp, "video_sink");
        }
    }

  if (!self->priv->sinkpad)
    self->priv->sinkpad = gst_bin_find_unconnected_pad (GST_BIN (bin),
        GST_PAD_SINK);

  if (!self->priv->sinkpad)
    {
      g_warning ("Could not find unconnected sink pad in the sin bin");
      gst_object_unref (bin);
      return NULL;
    }

  self->priv->caps_changed_id = g_signal_connect (self->priv->sinkpad,
      "notify::caps", G_CALLBACK (caps_changed), self);

  if (g_getenv ("STREAM_ENGINE_FPS"))
    gst_pad_add_event_probe (self->priv->sinkpad, G_CALLBACK (fps_qos_probe),
  self);

  ghostpad = gst_ghost_pad_new ("sink", self->priv->sinkpad);
  if (!gst_element_add_pad (bin, ghostpad))
    {
      gst_object_unref (ghostpad);
      g_warning ("Could not add sink ghostpad to the sink bin");
      gst_object_unref (bin);
      return NULL;
    }

  sink = bin;
  return sink;
}

#if 0
static GstElement *
make_video_sink (TpStreamEngineVideoSink *self, gboolean is_preview)
{
  const gchar *videosink_name;
  GstElement *sink = NULL;
  GstElement *bin, *tmp;
  GstPad *ghostpad;

  if ((videosink_name = getenv ("PREVIEW_VIDEO_SINK")) ||
      (videosink_name = getenv ("FS_VIDEO_SINK")) ||
      (videosink_name = getenv ("FS_VIDEOSINK")))
    {
      g_debug ("making video sink with pipeline \"%s\"", videosink_name);
      sink = gst_parse_bin_from_description (videosink_name, TRUE, NULL);
    }
  else
    {
      sink = gst_element_factory_make ("xvimagesink", NULL);
    }

  if (sink == NULL)
    {
      g_debug ("failed to make a video sink");
      return NULL;
    }

  if (g_object_has_property (G_OBJECT (sink), "force-aspect-ratio"))
    g_object_set (sink, "force-aspect-ratio", FALSE, NULL);

  if (!g_object_has_property (G_OBJECT (sink), "colorkey"))
    g_warning ("Sink does not have a colorkey property, UI may be wrong");

  g_object_set (sink,
      "colorkey", 0xff00ff,
      "autopaint-colorkey", FALSE,
      NULL);

  if (g_object_has_property (sink, "stacking"))
    g_object_set (sink, "stacking", is_preview ? 2 : 1, NULL);

  g_debug ("made video sink element %s", GST_ELEMENT_NAME (sink));

  bin = gst_bin_new (NULL);

  if (!gst_bin_add (GST_BIN (bin), sink))
    {
      g_warning ("Could not add source bin to the pipeline");
      gst_object_unref (bin);
      gst_object_unref (sink);
      return NULL;
    }

  if (g_getenv ("STREAM_ENGINE_FPS") &&
      !strcmp (g_getenv ("STREAM_ENGINE_FPS"), "overlay"))
    {
      tmp = gst_element_factory_make ("textoverlay", NULL);

      if (tmp != NULL)
        {
          self->priv->textoverlay = tmp;
          g_debug ("linking textoverlay");
          if (!gst_bin_add (GST_BIN (bin), tmp))
            {
              g_warning ("Could not add textoverlay to the source bin");
              gst_object_unref (tmp);
              gst_object_unref (bin);
              return NULL;
            }
          if (!gst_element_link (tmp, sink))
            {
              g_warning ("Could not link sink and textoverlay elements");
              gst_object_unref (bin);
              return NULL;
            }
          sink = tmp;
          self->priv->sinkpad = gst_element_get_static_pad (tmp, "video_sink");
        }
    }

  if (!self->priv->sinkpad)
    self->priv->sinkpad = gst_bin_find_unconnected_pad (GST_BIN (bin),
        GST_PAD_SINK);

  if (!self->priv->sinkpad)
    {
      g_warning ("Could not find unconnected sink pad in the source bin");
      gst_object_unref (bin);
      return NULL;
    }

  self->priv->caps_changed_id = g_signal_connect (self->priv->sinkpad,
      "notify::caps", G_CALLBACK (caps_changed), self);

  if (g_getenv ("STREAM_ENGINE_FPS"))
    gst_pad_add_event_probe (self->priv->sinkpad, G_CALLBACK (fps_qos_probe),
  self);

  ghostpad = gst_ghost_pad_new ("sink", self->priv->sinkpad);
  if (!gst_element_add_pad (bin, ghostpad))
    {
      gst_object_unref (ghostpad);
      g_warning ("Could not add sink ghostpad to the source bin");
      gst_object_unref (bin);
      return NULL;
    }

  sink = bin;

  return sink;
}
#endif

static GObject *
tp_stream_engine_video_sink_constructor (GType type,
    guint n_props,
    GObjectConstructParam *props)
{
  GObject *obj;
  TpStreamEngineVideoSink *self = NULL;
  g_debug("video_sink_constructor");

  obj = G_OBJECT_CLASS (tp_stream_engine_video_sink_parent_class)->constructor (type, n_props, props);

  self = (TpStreamEngineVideoSink *) obj;

  self->priv->sink = make_video_sink (self, self->priv->is_preview);

  if (self->priv->sink)
    gst_object_ref (self->priv->sink);

  return obj;
}

static void
tp_stream_engine_video_sink_dispose (GObject *object)
{
  TpStreamEngineVideoSink *self = TP_STREAM_ENGINE_VIDEO_SINK (object);
  close_ipc (self);
  if (self->priv->sinkpad)
    {
      if (self->priv->caps_changed_id)
        g_signal_handler_disconnect (self->priv->sinkpad,
            self->priv->caps_changed_id);
      self->priv->caps_changed_id = 0;

      gst_object_unref (self->priv->sinkpad);
      self->priv->sinkpad = NULL;
    }

  if (self->priv->sink)
    {
      gst_object_unref (self->priv->sink);
      self->priv->sink = NULL;
    }

  g_static_mutex_lock (&self->priv->mutex);
  if (self->priv->size_change_idle_id)
    g_source_remove (self->priv->size_change_idle_id);
  self->priv->size_change_idle_id = 0;
  g_static_mutex_unlock (&self->priv->mutex);

  if (G_OBJECT_CLASS (tp_stream_engine_video_sink_parent_class)->dispose)
    G_OBJECT_CLASS (tp_stream_engine_video_sink_parent_class)->dispose (
        object);
}


static GstElement *
tp_stream_engine_video_sink_get_xv_device (TpStreamEngineVideoSink *self)
{
  GstIterator *iter = gst_bin_iterate_all_by_interface (
      GST_BIN (self->priv->sink),
      GST_TYPE_X_OVERLAY);
  GstElement *sink = NULL;
  gpointer item;
  gboolean done = FALSE;

  while (done == FALSE)
    {
      switch (gst_iterator_next (iter, &item))
        {
        case GST_ITERATOR_OK:
          if (g_object_has_property (G_OBJECT (item), "device"))
            {
              done = TRUE;
              sink = GST_ELEMENT (item);
            }
          else
            {
              gst_object_unref (item);
            }
          break;
        case GST_ITERATOR_RESYNC:
          if (sink)
            gst_object_unref (sink);
          sink = NULL;
          gst_iterator_resync (iter);
          break;
        case GST_ITERATOR_ERROR:
          done = TRUE;
          break;
        case GST_ITERATOR_DONE:
          done = TRUE;
          break;
        }
    }

  gst_iterator_free (iter);

  return sink;
}

static void
do_nothing_cb (GstPad *pad, gboolean blocked, gpointer data)
{
}

static void
restart_sink_pad_blocked (GstPad *pad, gboolean blocked, gpointer data)
{
  GstPad *peerpad;
  GstElement *sink;

  peerpad = gst_pad_get_peer (pad);
  g_assert (peerpad);

  sink = gst_pad_get_parent_element (peerpad);

  gst_element_set_state (sink, GST_STATE_NULL);
  gst_element_sync_state_with_parent (sink);

  gst_object_unref (sink);
  gst_object_unref (peerpad);
  gst_pad_set_blocked_async (pad, FALSE, do_nothing_cb, NULL);
}

static void
tp_stream_engine_video_sink_set_property  (GObject *object,
    guint property_id,
    const GValue *value,
    GParamSpec *pspec)
{
  TpStreamEngineVideoSink *self = TP_STREAM_ENGINE_VIDEO_SINK (object);

    switch (property_id)
    {
    case PROP_IS_PREVIEW:
      self->priv->is_preview = g_value_get_boolean (value);
      break;
    case PROP_XV_DEVICE:
      {
        GstElement *xvdevice =
            tp_stream_engine_video_sink_get_xv_device (self);
        gchar *current_dev;

        if (!xvdevice)
          return;
        g_object_get (xvdevice, "device", &current_dev, NULL);
        if (strcmp (current_dev, g_value_get_string (value)))
          {
            GstPad *sinkpad;
            GstPad *srcpad;
            g_object_set_property (G_OBJECT (xvdevice), "device", value);

            sinkpad = gst_element_get_static_pad (xvdevice, "sink");
            g_assert (sinkpad);
            srcpad = gst_pad_get_peer (sinkpad);
            g_assert (sinkpad);
            gst_object_unref (sinkpad);

            gst_pad_set_blocked_async (srcpad, TRUE, restart_sink_pad_blocked, self);
            gst_object_unref (srcpad);
          }
        gst_object_unref (xvdevice);
      }
      break;

    case PROP_WINDOW_ID:
      self->priv->window_id = g_value_get_uint (value);
      g_debug ("Set window id %u for video_sink %s\n",
          self->priv->window_id,
          GST_ELEMENT_NAME (self->priv->sink));
      break;

    case PROP_WINDOW_PARAMS:
      {
        WindowParams *params = g_value_get_pointer (value);
        g_debug ("Set window params for video_sink %s\n", GST_ELEMENT_NAME (self->priv->sink));
        g_debug ("Window Params x:%u y:%u width:%u height:%u \n",
            params->x,
            params->y,
            params->width,
            params->height);
        self->priv->win_params.x = params->x;
        self->priv->win_params.y = params->y;
        self->priv->win_params.width = params->width;
        self->priv->win_params.height = params->height;
        break;
      }
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
tp_stream_engine_video_sink_get_property  (GObject *object,
    guint property_id,
    GValue *value,
    GParamSpec *pspec)
{
  TpStreamEngineVideoSink *self = TP_STREAM_ENGINE_VIDEO_SINK (object);

    switch (property_id)
    {
    case PROP_SINK:
      g_value_set_object (value, self->priv->sink);
      break;
    case PROP_WINDOW_ID:
      g_value_set_uint (value, self->priv->window_id);
      break;
    case PROP_XV_DEVICE:
      {
        GstElement *xvdevice =
            tp_stream_engine_video_sink_get_xv_device (self);

        if (!xvdevice)
          return;
        g_object_get_property (G_OBJECT (xvdevice), "device", value);
        gst_object_unref (xvdevice);
      }
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}


static void
tp_stream_engine_video_sink_class_init (TpStreamEngineVideoSinkClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (TpStreamEngineVideoSinkPrivate));
  object_class->dispose = tp_stream_engine_video_sink_dispose;
  object_class->constructor = tp_stream_engine_video_sink_constructor;
  object_class->set_property = tp_stream_engine_video_sink_set_property;
  object_class->get_property = tp_stream_engine_video_sink_get_property;

  g_object_class_install_property (object_class, PROP_SINK,
      g_param_spec_object ("sink",
          "The video sink element",
          "The GstElement that is used as videosink",
          GST_TYPE_ELEMENT,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (object_class, PROP_WINDOW_ID,
      g_param_spec_uint ("window-id",
          "Window id",
          "The window id to Xembed",
          0, G_MAXULONG, 0,
          G_PARAM_READWRITE));

  g_object_class_install_property (object_class, PROP_WINDOW_PARAMS,
      g_param_spec_pointer ("window-params",
          "Window parameters", 
          "The window paramters set from the UI for the video windows",
          G_PARAM_READWRITE));

  g_object_class_install_property (object_class, PROP_IS_PREVIEW,
      g_param_spec_boolean ("is-preview",
          "Window id",
          "The window id to Xembed",
          FALSE,
          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (object_class, PROP_XV_DEVICE,
      g_param_spec_string ("xv-device",
          "Xv Device",
          "The Xv Device to use",
          DEFAULT_XV_DEVICE,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  signals[PLUG_DELETED] =
      g_signal_new ("plug-deleted",
          G_OBJECT_CLASS_TYPE (klass),
          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
          0,
          NULL, NULL,
          g_cclosure_marshal_VOID__VOID,
          G_TYPE_NONE, 0);

  signals[SIZE_CHANGED] =
      g_signal_new ("size-changed",
          G_OBJECT_CLASS_TYPE (klass),
          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
          0,
          NULL, NULL,
          tp_stream_engine_marshal_VOID__UINT_UINT,
          G_TYPE_NONE, 2, G_TYPE_UINT, G_TYPE_UINT);
}

struct xid_data
{
  GstElement *src;
  gulong window_id;
  WindowParams win_params;
  gboolean found;
};

static void
set_window_xid (gpointer data, gpointer user_data)
{
  GstXOverlay *xov = GST_X_OVERLAY (data);
  struct xid_data *xiddata = (struct xid_data *) user_data;

  if (GST_ELEMENT_CAST(xov) == xiddata->src) {
      if (xiddata->window_id == 0)
      {
        g_debug ("No window id and params specified, will open a new window,\
            instead of crashing");
        return;
      }    
      gst_x_overlay_set_xwindow_id (xov, xiddata->window_id);
      gst_x_overlay_set_render_rectangle (xov,
          xiddata->win_params.x,
          xiddata->win_params.y,
          xiddata->win_params.width,
          xiddata->win_params.height);
          xiddata->found = TRUE;
  }
  g_debug("Setwindowxid, window id= %d x=%d y=%d %d x %d ", xiddata->window_id,
          xiddata->win_params.x,
          xiddata->win_params.y,
          xiddata->win_params.width,
          xiddata->win_params.height);

  gst_object_unref (data);
}

gboolean
tp_stream_engine_video_sink_bus_sync_message (
    TpStreamEngineVideoSink *self,
    GstMessage *message)
{
  const GstStructure *s;
  GstIterator *it = NULL;
  struct xid_data xiddata;

  if (GST_MESSAGE_TYPE (message) != GST_MESSAGE_ELEMENT)
    return FALSE;

  s = gst_message_get_structure (message);
  if (!gst_structure_has_name (s, "prepare-xwindow-id"))
    return FALSE;

  xiddata.src = GST_ELEMENT (GST_MESSAGE_SRC (message));
  xiddata.window_id = self->priv->window_id;
  xiddata.win_params = self->priv->win_params;
  xiddata.found = FALSE;

  g_debug("In bus sync handler, window id= %d", self->priv->window_id);

  it = gst_bin_iterate_all_by_interface (GST_BIN (self->priv->sink),
      GST_TYPE_X_OVERLAY);
  while (gst_iterator_foreach (it, set_window_xid, &xiddata) ==
      GST_ITERATOR_RESYNC)
    gst_iterator_resync (it);
  gst_iterator_free (it);

  return xiddata.found;
}
